"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""


import pynta
import pyparamnta
from misc_functions import split_constraints 

import re
import copy
import logging

logging.basicConfig()
logger = logging.getLogger('MODEL_TRANSFORM')


def add_universal_location(template,outputs,inputs):
  """ Add a universal location to the templates:
      - The location is named "Universal". No location is added if the template already contains one with this name.
      - The location has self-loop transitions for each input and output action. """
  universal = template.get_location("Universal")
  if universal:
    logger.error("The template already has a location named \"Universal\"")
    return universal
  universal = pynta.Location(template,"Universal")
  template.locations.append(universal)
  for action in outputs:
    new_t = pynta.Transition(source=universal, \
				target = universal, \
				controllable = True, \
				synchronisation = (action,"!"))
    template.transitions.append(new_t)
    universal.forward.append(new_t)
  for action in inputs:
    new_t = pynta.Transition(source=universal, \
				target = universal, \
				controllable = True, \
				synchronisation = (action,"?"))
    template.transitions.append(new_t)
    universal.forward.append(new_t)
  return universal

def contraryConstraint(cons):
  """Return the contrary of a simple rectangular guard:
  e.g. x <= k --> x > k, where k is a positive integer."""
  if not cons:
    return ""
  assert(isinstance(cons,str))
  r = re.search('^\s*(?P<clock>[a-zA-Z_]\w*)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<coeff>\d+)\s*$',cons)
  if r:
    if r.group("sign") == "<":
      return r.group("clock") + '>=' + r.group('coeff')
    elif r.group("sign") == "<=":
      return r.group("clock") + '>' + r.group('coeff')
    elif r.group("sign") == ">":
      return r.group("clock") + '<=' + r.group('coeff')
    elif r.group("sign") == ">=":
      return r.group("clock") + '<' + r.group('coeff')
    else:
      return r.group("clock") + '<' + r.group('coeff') + ' && ' + r.group("clock") + '>' + r.group('coeff')
      
  r = re.search('^\s*(?P<coeff>\d+)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<clock>[a-zA-Z_]\w*)\s*$',cons)
  if r:
    if r.group("sign") == ">":
      return r.group("clock") + '>=' + r.group('coeff')
    elif r.group("sign") == ">=":
      return r.group("clock") + '>' + r.group('coeff')
    elif r.group("sign") == "<":
      return r.group("clock") + '<=' + r.group('coeff')
    elif r.group("sign") == "<=":
      return r.group("clock") + '<' + r.group('coeff')
    else:
      return r.group("clock") + '<' + r.group('coeff') + ' && ' + r.group("clock") + '>' + r.group('coeff')
  logger.error("can't get contrary constraint: "+cons)
  return ""
  

def enlargedConstraint(cons,delta):
  """Enlarge a simple rectangular constraint by delta:
  e.g. x <= k --> x <= k+delta (where delta is either an integer or a parameter).
  
  Return the enlarge constraint."""
  if not cons:
    return ""
  assert(isinstance(cons,str))
  assert(isinstance(delta,str) | isinstance(delta,int))
  r = re.search('^\s*(?P<clock>[a-zA-Z_]\w*)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<coeff>\d+)\s*$',cons)
  if r:
    if r.group("sign") == "<":
      if isinstance(delta,int):
	return r.group("clock") + '<' + str(int(r.group('coeff'))+delta)
      else:
	return r.group("clock") + '<' + r.group('coeff') +'+'+ str(delta)
    elif r.group("sign") == "<=":
      if isinstance(delta,int):
	return r.group("clock") + '<=' + str(int(r.group('coeff'))+delta)
      else:
	return r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta)
    elif r.group("sign") == ">":
      if isinstance(delta,int):
	return r.group("clock") + '>' + str(int(r.group('coeff'))-delta)
      else:
	return r.group("clock") + '>' + r.group('coeff') +'-'+ str(delta)
    elif r.group("sign") == ">=":
      if isinstance(delta,int):
	return r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)
      else:
	return r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)
    else: # sign is == : The constraint is split into 2 inequalities.
      if isinstance(delta,int):
	return r.group("clock") + '<=' + str(int(r.group('coeff'))+delta) + ' && ' + r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)	  
      else:
	return r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta) + ' && ' + r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)
	
  r = re.search('^\s*(?P<coeff>\d+)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<clock>[a-zA-Z_]\w*)\s*$',cons)
  if r:
    if r.group("sign") == ">":
      if isinstance(delta,int):
	return r.group("clock") + '<' + str(int(r.group('coeff'))+delta)
      else:
	return r.group("clock") + '<' + r.group('coeff') +'+'+ str(delta)
    elif r.group("sign") == ">=":
      if isinstance(delta,int):
	return r.group("clock") + '<=' + str(int(r.group('coeff'))+delta)
      else:
	return r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta)
    elif r.group("sign") == "<":
      if isinstance(delta,int):
	return r.group("clock") + '>' + str(int(r.group('coeff'))-delta)
      else:
	return r.group("clock") + '>' + r.group('coeff') +'-'+ str(delta)
    elif r.group("sign") == "<=":
      if isinstance(delta,int):
	return r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)
      else:
	return r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)
    else: # sign is == : The constraint is split into 2 inequalities.
      if isinstance(delta,int):
	return r.group("clock") + '<=' + str(int(r.group('coeff'))+delta) + ' && ' + r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)	  
      else:
	return r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta) + ' && ' + r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)
  logger.error("can't enlarge constraint: "+cons)
  return ""
  
  
def restrictedConstraint(cons,delta):
  """Reduced a simple rectangular constraint by delta:
  e.g. x <= k --> x <= k-delta (where delta is either an integer or a parameter).
  Return a list with the reduced constraint and its contrary. 
  If equality return an empty list."""
  if not cons:
    return []
  assert(isinstance(cons,str))
  assert(isinstance(delta,str) | isinstance(delta,int))
  r = re.search('^\s*(?P<clock>[a-zA-Z_]\w*)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<coeff>\d+)\s*$',cons)
  if r:
    if r.group("sign") == "<":
      if isinstance(delta,int):
	return [r.group("clock") + '<' + str(int(r.group('coeff'))-delta), r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)]
      else:
	return [r.group("clock") + '<' + r.group('coeff') +'-'+ str(delta), r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)]
    elif r.group("sign") == "<=":
      if isinstance(delta,int):
	return [r.group("clock") + '<=' + str(int(r.group('coeff'))-delta), r.group("clock") + '>' + str(int(r.group('coeff'))-delta)]
      else:
	return [r.group("clock") + '<=' + r.group('coeff') +'-'+ str(delta), r.group("clock") + '>' + r.group('coeff') +'-'+ str(delta)]
    elif r.group("sign") == ">":
      if isinstance(delta,int):
	return [r.group("clock") + '>' + str(int(r.group('coeff'))+delta), r.group("clock") + '<=' + str(int(r.group('coeff'))+delta)]
      else:
	return [r.group("clock") + '>' + r.group('coeff') +'+'+ str(delta), r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta)]
    elif r.group("sign") == ">=":
      if isinstance(delta,int):
	return [r.group("clock") + '>=' + str(int(r.group('coeff'))+delta), r.group("clock") + '<' + str(int(r.group('coeff'))+delta)]
      else:
	return [r.group("clock") + '>=' + r.group('coeff') +'+'+ str(delta), r.group("clock") + '<' + r.group('coeff') +'+'+ str(delta)]
    else:
      return []
  r = re.search('^\s*(?P<coeff>\d+)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<clock>[a-zA-Z_]\w*)\s*$',cons)
  if r:
    if r.group("sign") == ">":
      if isinstance(delta,int):
	return [r.group("clock") + '<' + str(int(r.group('coeff'))-delta), r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)]
      else:
	return [r.group("clock") + '<' + r.group('coeff') +'-'+ str(delta), r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)]
    elif r.group("sign") == ">=":
      if isinstance(delta,int):
	return [r.group("clock") + '<=' + str(int(r.group('coeff'))-delta), r.group("clock") + '>' + str(int(r.group('coeff'))-delta)]
      else:
	return [r.group("clock") + '<=' + r.group('coeff') +'-'+ str(delta), r.group("clock") + '>' + r.group('coeff') +'-'+ str(delta)]
    elif r.group("sign") == "<":
      if isinstance(delta,int):
	return [r.group("clock") + '>' + str(int(r.group('coeff'))+delta), r.group("clock") + '<=' + str(int(r.group('coeff'))+delta)]
      else:
	return [r.group("clock") + '>' + r.group('coeff') +'+'+ str(delta), r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta)]
    elif r.group("sign") == "<=":
      if isinstance(delta,int):
	return [r.group("clock") + '>=' + str(int(r.group('coeff'))+delta), r.group("clock") + '<' + str(int(r.group('coeff'))+delta)]
      else:
	return [r.group("clock") + '>=' + r.group('coeff') +'+'+ str(delta), r.group("clock") + '<' + r.group('coeff') +'+'+ str(delta)]
    else:
      return []
  logger.error("can't restrict constraint: "+cons)
  return []

def scaledConstraint(cons,factor):
  """Scaled the constant coefficient in a simple guard by a factor:
  e.g. x <= k --> x <= k*factor
        x <= k + l*d --> x <= k*factor + l*d """
  if not cons:
    return ""
  assert(isinstance(cons,str))
  assert(isinstance(factor,int))

  r = re.search('^\s*(?P<clock>[a-zA-Z_]\w*)\s*(?P<sign1>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<coeff1>(-)?\s*\d+)?\s*(?P<sign2>((-)|(\+)))?\s*(?P<coeff2>\d+)?\s*(?P<sign3>(\*))?\s*(?P<delta>\w+)?\s*$',cons)
  coeff_str=""
  if r:
    if r.group("coeff1"):
      if not r.group("delta") or r.group("sign2"):
	coeff = int(r.group("coeff1"))*factor
	coeff_str +=  str(coeff)
      else:
	coeff = int(r.group("coeff1"))
	coeff_str += str(coeff)
    if r.group("sign2"):
      coeff_str += r.group("sign2")
    if r.group("coeff2"):
      coeff = int(r.group("coeff2"))
      coeff_str += str(coeff)
    if r.group("sign3"):
      coeff_str += r.group("sign3")
    if r.group("delta"):
      coeff_str += r.group("delta")
    return r.group("clock") + r.group("sign1") + coeff_str
    
  r = re.search('^\s*(?P<coeff1>(-)?\s*\d+)?\s*(?P<sign2>((-)|(\+)))?\s*(?P<coeff2>\d+)?\s*(?P<sign3>(\*))?\s*(?P<delta>\w+)?\s*(?P<sign1>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<clock>[a-zA-Z_]\w*)\s*$',cons)
  if r:
    if r.group("coeff1"):
      if not r.group("delta") or r.group("sign2"):
	coeff = int(r.group("coeff1"))*factor
	coeff_str += str(coeff)
      else:
	coeff = int(r.group("coeff1"))
	coeff_str += str(coeff)
    if r.group("sign2"):
      coeff_str += r.group("sign2")
    if r.group("coeff2"):
      coeff = int(r.group("coeff2"))
      coeff_str += str(coeff)
    if r.group("sign3"):
      coeff_str += r.group("sign3")
    if r.group("delta"):
      coeff_str += r.group("delta")
    return coeff_str + r.group("sign1") + r.group("clock")
  logger.error("can't scale constraint: "+cons)
  return cons
 

def scaledTemplateConstants(template,factor):
  """Scale the constants coefficient in a rectangular automata.  
  Return a new modified template."""
  assert(isinstance(factor,int))
  assert(isinstance(template,pynta.Template))
  
  scaled_template = copy.deepcopy(template)
  scaled_template.name += "S"+str(factor)
  
  for l in scaled_template.locations:
    l.template = scaled_template
    invl = split_constraints(l.invariant)
    for i in range(len(invl)):
      invl[i] = scaledConstraint(invl[i],factor)
    l.invariant = ' && '.join(invl)
      
  for t in scaled_template.transitions:
    t.template = scaled_template
    guardl = split_constraints(t.guard)
    for i in range(len(guardl)):
      guardl[i] = scaledConstraint(guardl[i],factor)
    t.guard = ' && '.join(guardl)
  
  return scaled_template

#def scaledTemplateConstants(template,factor):
  #"""Scale the constants coefficient in a rectangular automata.  
  #Return a new modified template."""
  #assert(isinstance(factor,int))
  #assert(isinstance(template,pynta.Template))
  
  #scaled_template = copy.deepcopy(template)
  #scaled_template.name += "S"+str(factor)

  #for l in scaled_template.locations:
    #l.template = scaled_template
    #invariant_str = str(l.invariant_dbm)
    #if invariant_str == "true":
      #l.invariant = l.invariant_var
    #else:
      #invl = invariant_str[1:-1].split("&")
      #for i in range(len(invl)):
	#invl[i] = scaledConstraint(invl[i],factor)
      #l.invariant = ' && '.join(invl)
      #if l.invariant_var:
	#l.invariant += " and " +  l.invariant_var
      
  #for t in scaled_template.transitions:
    #t.template = scaled_template
    #guard_str = str(t.guard_dbm)
    #if guard_str == "true":
      #t.guard = t.guard_var
    #else:
      #guardl = guard_str[1:-1].split("&")
      #for i in range(len(guardl)):
	#guardl[i] = scaledConstraint(guardl[i],factor)
      #t.guard = ' && '.join(guardl)
      #if t.guard_var:
	#t.guard += " and " +  t.guard_var    
  #return scaled_template
    

def implemPerturbation(template,delta,outputs,inputs,param=0):
  """Add a delay perturbation delta to the guards of the template,
  (where delta is either an integer or a parameter).
  Return a new template for the perturbed implementation."""
  if isinstance(delta,str):
    rob_impl = pyparamnta.ParamTemplate(template,delta,param)
  else:
    rob_impl = copy.deepcopy(template)
  rob_impl.name += "Rob"+str(delta)
  
  universal = add_universal_location(rob_impl,outputs,inputs)
  
  for l in rob_impl.locations:
    l.template = rob_impl
    invariant_str = str(l.invariant_dbm)
    if invariant_str == "true":
      l.invariant = l.invariant_var
    else:
      invl = invariant_str[1:-1].split("&")
      for i in range(len(invl)):
	invl[i] = enlargedConstraint(invl[i],delta)
      l.invariant = ' && '.join(invl)
      if l.invariant_var:
	l.invariant += " and " +  l.invariant_var
      
  for t in rob_impl.transitions:
    t.template = rob_impl
    if t.controllable:  	# output transition
      guard_str = str(t.guard_dbm)
      if guard_str == "true":
	t.guard = t.guard_var
      else:
	guardl = guard_str[1:-1].split("&")
	for i in range(len(guardl)):
	  guardl[i] = enlargedConstraint(guardl[i],delta)
	t.guard = ' && '.join(guardl)
	if t.guard_var:
	  t.guard += " and " +  t.guard_var
    else:			# input transition
      guard_str = str(t.guard_dbm)
      if guard_str != "true":
	guardl = guard_str[1:-1].split("&")
	guardl_rob = []
	for i in range(len(guardl)):
	  cons_robl = restrictedConstraint(guardl[i],delta)
	  if cons_robl:
	    guardl_rob.append(cons_robl[0])
	    guard_lu=guardl[:]
	    guard_lu.append(cons_robl[1])
	    guard_lu = ' && '.join(guard_lu)
	    if t.guard_var:
	      guard_lu += " and " + t.guard_var
	    lu_transitions.append(pynta.Transition(source=t.source, \
			  target = universal, \
			  synchronisation = t.synchronisation, \
			  guard = guard_lu))
	t.guard = ' && '.join(guardl_rob)
	if t.guard_var:
	  t.guard += " and " +  t.guard_var      
     
  rob_impl.transitions.extend(lu_transitions)
  #makeInputEnabled(rob_impl,inputs,universal)
  rob_impl.update_forward_backward()
  return rob_impl
  

  
def robustTGA(template,delta,bad_loc_name="BAD",param=0,):
  """Compute the robust TGA in which one player wins iff it has a delta-robust strategy.
  Delta can be either a value or parameter.
  Controllable transitions must be assigned to the desired player.
  Return a new template for the robust TGA."""
  if isinstance(delta,str):
    rob_game = pyparamnta.ParamTemplate(template,delta,param)
  else:
    rob_game = copy.deepcopy(template)
  rob_game.name += "RobTG"
  rob_game.clocks["robClk"] = None
  
  bad_loc = None
  for l in rob_game.locations:
    l.template = rob_game
    if str(l.name) == bad_loc_name:
      bad_loc = l
      break;
  if not bad_loc:
    bad_loc = pynta.Location(rob_game,name=bad_loc_name)
    rob_game.locations.append(bad_loc)
  
  new_transitions = []
  removed_transitions = []
  
  for t in rob_game.transitions:
    t.template = rob_game
    if t.controllable:
      guard_str = str(t.guard_dbm)
      if re.search('(==)',guard_str):
	removed_transitions.append(t) # if the guard is an equality the transition is permanently removed since it cannot be robustly taken
        continue
	
      source_loc = t.source
      target_loc = t.target
      alpha_loc = pynta.Location(rob_game,name=str(source_loc.name)+"_"+t.synchronisation[0]+"_"+str(target_loc.name)+"_A")
      beta_loc = pynta.Location(rob_game,name=str(source_loc.name)+"_"+t.synchronisation[0]+"_"+str(target_loc.name)+"_B")
      alpha_loc.invariant = "robClk <= "+str(delta)
      rob_game.locations.append(alpha_loc)
      rob_game.locations.append(beta_loc)
      
      if guard_str != "true":
	guard_str = guard_str[1:-1]
	guardl = guard_str.split("&")
	inv_str = str(source_loc.invariant_dbm)
	if inv_str != "true":
	  inv_str = inv_str[1:-1]
	  guardl.extend(inv_str.split("&"))
	for guardi in guardl:
	  if re.search("<=|<",guardi):
	    guard = contraryConstraint(guardi)+" && robClk <= "+str(delta)
	    if t.guard_var:
	      guard += " and " +  t.guard_var   
	    new_transitions.append(pynta.Transition(source=alpha_loc, \
		  target = bad_loc, \
		  controllable = False, \
		  guard = guard, \
		  update="robClk=0"))
	    new_transitions.append(pynta.Transition(source=beta_loc, \
		  target = bad_loc, \
		  controllable = False, \
		  guard = guard, \
		  update="robClk=0"))
           
      tupdate = t.update
      if tupdate:
	tupdate += ",robClk=0"
      else:
	tupdate += "robClk=0"
	
      guard = "robClk == "+str(delta)
      if t.guard_var:
	guard += " and " +  t.guard_var 
      new_transitions.append(pynta.Transition(source=alpha_loc, \
	      target = beta_loc, \
	      controllable = True, \
	      guard = guard, \
	      update="robClk=0"))
      guard = "robClk <= "+str(delta)
      if t.guard_var:
	guard += " and " +  t.guard_var 
      new_transitions.append(pynta.Transition(source=alpha_loc, \
	      target = target_loc, \
	      controllable = False, \
	      guard = guard, \
	      update=tupdate))
	      
      new_transitions.append(pynta.Transition(source=beta_loc, \
	      target = target_loc, \
	      controllable = False, \
	      guard = guard, \
	      update=tupdate))
      
      t.target = alpha_loc
      t.update="robClk=0"
      
      for u in rob_game.transitions:
	if not u.controllable and u.source==source_loc and u.target != u.source: #self-loop with input sync are not added to the robust moved of the spoiler
	  uguard = str(u.guard_dbm)
	  if uguard != "true":
	    uguard = uguard[1:-1]
	    uguard += " && robClk <= "+str(delta)
	    if u.guard_var:
	      uguard += " and " +  u.guard_var 
	  else:
	    uguard = "robClk <= "+str(delta)
	    if u.guard_var:
	      uguard += " and " +  u.guard_var 
	  uupdate = str(u.update)
	  if uupdate:
	    if not re.search("robClk=0",uupdate):
	      uupdate += ",robClk=0"
	  else:
	    uupdate = "robClk=0"
	  found = False
	  if u.target == bad_loc:
	    for t in new_transitions: #we search for equivalent transitions, which happen specifically if we apply the transformations: consistencyGame and robustTGA successively.  
	      if (t.source == alpha_loc or t.source == beta_loc) and \
		  t.target == bad_loc and \
		  t.guard == uguard and \
		  t.update == uupdate:
		found = True
		break
	  if not found:  
	    new_transitions.append(pynta.Transition(source=alpha_loc, \
		    target = u.target, \
		    controllable = False, \
		    guard = uguard, \
		    synchronisation = u.synchronisation, \
		    update=uupdate))
	    
	    new_transitions.append(pynta.Transition(source=beta_loc, \
		    target = u.target, \
		    controllable = False, \
		    guard = uguard, \
		    synchronisation = u.synchronisation, \
		    update=uupdate))
    else:
      if t.target != t.source:
	if t.update:
	  t.update += ", robClk=0"
	else:
	  t.update = "robClk=0"
      
  rob_game.transitions.extend(new_transitions)
  for t in removed_transitions:
    rob_game.transitions.remove(t)
  rob_game.update_forward_backward()   
  return rob_game 
  

def consistencyGame(template,delta=0):
  """Prepare a specification for solving the consistency game.
  - indepenent progress condition: 
    - Remove invariants and replace then by input edges whose target is an inconsistent location.
    - Restrict the guards by the invariants: on outcome edges, and on income edges if the clock is not reset
    (invariants are assumed to be only greater bound; clocks are only reset to zero).
  - delta latency of inputs (if delta != 0):
    - increased guard on inputs by delta
    - add edges to inconsistent location for each pair of edges that start from the same location, with the same input synchronisation
    - add as a guard to these edges the conjunction of the two increased guards.
  - Check that output actions are controllable and input non controllable.
  """
  cons_game = copy.deepcopy(template)
  cons_game.name += "ConsTG"
  for l in cons_game.locations:
    l.template = cons_game
  for t in cons_game.transitions:
    t.template = cons_game
  
  incons_loc = pynta.Location(cons_game,name="Inconsistent")
  cons_game.locations.append(incons_loc)
  
  if delta:
    for t in cons_game.transitions:
      if t.synchronisation[1] == '?' and t.target.name != "Universal":
	guard_str = str(t.guard_dbm)
	if guard_str == "true":
	  t.guard = t.guard_var
	else:
	  guard_str = guard_str[1:-1]
	  guardl = guard_str.split("&")
	  for i in range(len(guardl)):
	    guardl[i] = enlargedConstraint(guardl[i],delta)
	  t.guard = ' && '.join(guardl)
	  if t.guard_var:
	    t.guard += " and " +  t.guard_var
    for l in cons_game.locations:
      i=0
      length = len(cons_game.transitions)
      for ti in range(length):
	ti = cons_game.transitions[i]
	if ti.source == l and  ti.target.name != "Universal" and  ti.synchronisation[1] == '?':
	  for j in range(i+1,length):
	    tj = cons_game.transitions[j]
	    if tj.source == l and  tj.target.name != "Universal" and tj.synchronisation == ti.synchronisation:
	      cons_game.transitions.append(pynta.Transition(source=l, \
		  target = incons_loc, \
		  controllable = False, \
		  synchronisation = tj.synchronisation, \
		  guard = ti.guard + " and " + tj.guard))
	    j+=1
	i+=1

  for t in cons_game.transitions:
    inv_str = str(t.source.invariant_dbm)
    if inv_str != "true":
      if t.guard:
	t.guard += " && " + inv_str[1:-1]
      else:
	t.guard = inv_str[1:-1]
    inv_str = str(t.target.invariant_dbm)	
    if t.source != t.target and inv_str != "true":
      updatedClk = []
      guard_str = str(t.guard_dbm)
      if guard_str != "true":
	guardl = guard_str[1:-1].split("&")
      else:
	guardl=[]
      for c in t.update_clock:
	updatedClk += [c.name]
      invl = inv_str[1:-1].split("&&")
      for invi in invl:
	r = re.search('([a-zA-Z_]\w*)', invi)
	invClk = r.group(1)
	if not invClk in updatedClk:
	  guardl.append(invi)
      t.guard = ' && '.join(guardl)
      if t.guard_var:
	t.guard += " and " +  t.guard_var
    if t.synchronisation[1] == '!':
      t.controllable = True
    else:
      t.controllable = False
    
  for l in cons_game.locations:
    if l != incons_loc:
      inv_str = str(l.invariant_dbm)	
      if inv_str != "true":
	invl = inv_str[1:-1].split("&")
	for invi in invl:
	    cons_game.transitions.append(pynta.Transition(source=l, \
		  target = incons_loc, \
		  controllable = False, \
		  guard = contraryConstraint(invi)))
	l.invariant = l.invariant_var
  cons_game.update_forward_backward()  
  return cons_game
  

def usefulnessGame(template,bad_loc_list=[]):
  """Prepare a specification for solving the usefulness game.
  - Remove invariants and add it to the guards of incoming and outcoming edges.
    If invariants are overpassed the environnement wins. Thus, contrary to the consistency game the spoiler (output player) is force to play.
  - Check that output actions are non controllable and input controllable.
  - all bad locations in bad_loc_list are renamed UNIV, such that there remains only one objective
  - remove self loop on inputs
  """
  useful_game = copy.deepcopy(template)
  useful_game.name += "UsefulTG"
  for l in useful_game.locations:
    l.template = useful_game
  for t in useful_game.transitions:
    t.template = useful_game

  #removed_transitions = []
  for t in useful_game.transitions:
    #if str(t.synchronisation)[-1] == '?' and t.source == t.target and str(t.assignment) == "":
      #removed_transitions.append(t)
      #continue
    inv_str = str(t.source.invariant_dbm)
    guard_str = str(t.guard_dbm)
    if inv_str != "true":
      if t.guard:
	t.guard += " && " + inv_str[1:-1]
      else:
	t.guard = inv_str[1:-1]
    inv_str = str(t.target.invariant_dbm)
    if t.source != t.target and inv_str != "true":
      updatedClk = []
      for c in t.update_clock:
	updatedClk += [c.name]
      if guard_str != "true":
	guardl = guard_str[1:-1].split("&")
      else:
	guardl=[]
      invl = inv_str[1:-1].split("&")
      for invi in invl:
	r = re.search('([a-zA-Z_]\w*)', invi)
	invClk = r.group(1)
	if not invClk in updatedClk:
	  guardl.append(invi)
      t.guard = ' && '.join(guardl)
      if t.guard_var:
	t.guard += " and " +  t.guard_var
    if t.synchronisation[1] == '!':
      t.controllable = False
    else:
      t.controllable = True
   
  for l in useful_game.locations:
    if l.name in bad_loc_list:
      l.name =  "Universal"
    l.invariant = l.invariant_var
  
  #for t in removed_transitions:
    #useful_game.transitions.remove(t)
  useful_game.update_forward_backward()  
  return useful_game  


def conjunction(spec1,spec2,outputs,inputs,universal1,universal2,merge_universal=False):
  """ Conjunction of two specifications sharing the same alphabet.
      The templates must have disjoint clocks, constants and variables.
  """
  if not (spec1.mgr is None and spec2.mgr is None):
    logger.error("Adaptive specifications cannot be conjunct.")
    return None
    
  conj = pynta.Template(spec1.nta,spec1.name+"AND"+spec2.name)
  for clock in spec1.clocks:
    conj.clocks[clock] = None
  for clock in spec2.clocks:
    conj.clocks[clock] = None
  conj.constants = dict(spec1.constants.items() + spec2.constants.items())
  conj.variables = dict(spec1.variables.items() + spec2.variables.items())

  universal = None
  if merge_universal:
    universal = add_universal_location(conj,outputs,inputs)
  
  loc_dict = {}
  for l1 in spec1.locations:
    if merge_universal and l1 == universal1:
      continue
    for l2 in spec2.locations:
      if merge_universal and l2 == universal2:
	continue
      l1_2 = pynta.Location(conj,name=str(l1.name) + "AND" + str(l2.name))
      conj.locations.append(l1_2)
      loc_dict[(l1,l2)]=l1_2
      if l1.invariant:
	if l2.invariant:
	  l1_2.invariant = l1.invariant + " and " + l2.invariant
	else:
	  l1_2.invariant = l1.invariant
      else:
	l1_2.invariant = l2.invariant
      if spec1.init_location == l1 and spec2.init_location == l2:
	conj.init_location = l1_2
	  
  for t1 in spec1.transitions:
    if merge_universal and t1.source == universal1:
      continue
    for t2 in spec2.transitions:
      if merge_universal and t2.source == universal2:
	continue
      if t1.synchronisation == t2.synchronisation: 	# equality between synchronisations: channel and channel type are equal
	assert(t1.controllable == t2.controllable)
	if merge_universal and (t1.target == universal1 or t2.target == universal2):
	  target = universal
	else:
	  target = loc_dict[(t1.target,t2.target)]
	t1_2 = pynta.Transition(source=loc_dict[(t1.source,t2.source)], \
			        target = target, \
				controllable = t1.controllable, \
				synchronisation = t1.synchronisation)
	if t1.guard:
	  if t2.guard:
	    t1_2.guard = t1.guard + " and " + t2.guard
	  else:
	    t1_2.guard = t1.guard
	else:
	  t1_2.guard = t2.guard
	if t1.update:
	  if t2.update:
	    t1_2.update = t1.update + " , " + t2.update
	  else:
	    t1_2.update = t1.update
	else:
	  t1_2.update = t2.update
	conj.transitions.append(t1_2)
  conj.update_forward_backward()	
  return conj
  

def composition(spec1,spec2,outputs1,outputs2,inputs1,inputs2,universal1,universal2,merge_universal=False):
  """ Parallel composition of two specifications.
      The templates must have disjoint clocks, constants and variables.
  """      
  if not (spec1.mgr is None and spec2.mgr is None):
    logger.error("Adaptive specifications cannot be composed.")
    return None
  compo = pynta.Template(spec1.nta,spec1.name+"PAR"+spec2.name)
  for clock in spec1.clocks:
    compo.clocks[clock] = None
  for clock in spec2.clocks:
    compo.clocks[clock] = None
  compo.constants = dict(spec1.constants.items() + spec2.constants.items())
  compo.variables = dict(spec1.variables.items() + spec2.variables.items())
  
  inputs = []
  for channel in inputs1:
    if channel not in outputs2:
      inputs += [channel]
  for channel in inputs2:
    if channel not in outputs1+inputs1:
      inputs += [channel]    
  
  universal = None
  if merge_universal:
    universal = add_universal_location(compo,outputs1+outputs2,inputs)
    
  loc_dict = {}
  for l1 in spec1.locations:
    if merge_universal and l1 == universal1:
      continue
    for l2 in spec2.locations:
      if merge_universal and l2 == universal2:
	continue
      l1_2 = pynta.Location(compo,name=str(l1.name)+"PAR"+str(l2.name))
      compo.locations.append(l1_2)
      loc_dict[(l1,l2)]=l1_2
      if l1.invariant:
	if l2.invariant:
	  l1_2.invariant = l1.invariant + " and " + l2.invariant
	else:
	  l1_2.invariant = l1.invariant
      else:
	l1_2.invariant = l2.invariant
      if spec1.init_location == l1 and spec2.init_location == l2:
	compo.init_location = l1_2

  for t1 in spec1.transitions:
    if merge_universal and t1.source == universal1:
      continue
    if t1.synchronisation[0] not in outputs2+inputs2:			 #t1 is not synchronized
      for l2 in spec2.locations:
	if merge_universal and l2 == universal2:
	  continue
	if merge_universal and (t1.target == universal1):
	  target = universal
	else:
	  target = loc_dict[(t1.target,l2)]
	t1_2 = pynta.Transition(source=loc_dict[(t1.source,l2)], \
				target = target, \
				controllable = t1.controllable, \
				guard = t1.guard, \
				synchronisation = t1.synchronisation, \
				update = t1.update)
	compo.transitions.append(t1_2)
    else:							#t1 is synchronized
      for t2 in spec2.transitions:
	if merge_universal and t2.source == universal2:
	  continue
	if t1.synchronisation[0] == t2.synchronisation[0]: 	# equality between the synchronisation channel
	  if merge_universal and (t1.target == universal1 or t2.target == universal2):
	    target = universal
	  else:
	    target = loc_dict[(t1.target,t2.target)]
	  t1_2 = pynta.Transition(source=loc_dict[(t1.source,t2.source)], \
				  target = target, \
				  controllable = t1.controllable or t2.controllable)
	  if t1.guard:
	    if t2.guard:
	      t1_2.guard = t1.guard + " and " + t2.guard
	    else:
	      t1_2.guard = t1.guard
	  else:
	    t1_2.guard = t2.guard
	  if t1.update:
	    if t2.update:
	      t1_2.update = t1.update + " , " + t2.update
	    else:
	      t1_2.update = t1.update
	  else:
	    t1_2.update = t2.update
	  if t1.synchronisation[1] ==  t2.synchronisation[1]:	# input-input synchronisation
	    t1_2.synchronisation = (t1.synchronisation[0],"?")
	  else:							# input-output synchronisation
	    t1_2.synchronisation = (t1.synchronisation[0],"!")
	  compo.transitions.append(t1_2)
  for t2 in spec2.transitions:
    if merge_universal and t2.source == universal2:
      continue
    if t2.synchronisation[0] not in outputs1+inputs1:			 #t2 is not synchronized
      for l1 in spec1.locations:
	if merge_universal and l1 == universal1:
	  continue
	if merge_universal and (t2.target == universal2):
	  target = universal
	else:
	  target = loc_dict[(l1,t2.target)]
	t1_2 = pynta.Transition(source=loc_dict[(l1,t2.source)], \
				target = target, \
				controllable = t2.controllable, \
				guard = t2.guard, \
				synchronisation = t2.synchronisation, \
				update = t2.update)
	compo.transitions.append(t1_2)
  compo.update_forward_backward()
  return compo
     

def quotient(spec1,spec2,outputs1,outputs2,inputs1,inputs2):
  """ Quotient of spec1 by spec2.
      The templates must have disjoint clocks, constants and variables.
  """      
  if not (spec1.mgr is None and spec2.mgr is None):
    logger.error("Adaptive specifications cannot be quotiented.")
    return None
  if spec1.nta != spec2.nta:
    logger.error("Specifications belongs to different networks.")
    return None
  if spec1.variables or spec2.variables:
    logger.error("Specifications with variables cannot be quotiented.")
    return None
  
  quotient = pynta.Template(spec1.nta,spec1.name+"DIV"+spec2.name)
  for quotient in spec1.clocks:
    quotient.clocks[clock] = None
  for clock in spec2.clocks:
    quotient.clocks[clock] = None
  quotient.constants = dict(spec1.constants.items() + spec2.constants.items())
   
  inputs = []
  inputs += inputs1
  for channel in output2:
    if channel not in inputs:
      inputs += [channel]
  outputs = []
  for channel in output1:
    if channel not in outputs2:
      outputs += [channel]
  
  universal = add_universal_location(quotient,outputs,inputs)
  inconsistent = pynta.Location(quotient,"Inconsistent",urgent=True)
  quotient.locations.append(inconsistent)
  
  loc_dict = {}
  for l1 in spec1.locations:
    for l2 in spec2.locations:
      l1_2 = pynta.Location(quotient,name=str(l1.name)+"DIV"+str(l2.name))
      quotient.locations.append(l1_2)
      loc_dict[(l1,l2)]=l1_2
      if spec1.init_location == l1 and spec2.init_location == l2:
	quotient.init_location = l1_2

  for l1 in spec1.locations:
    for l2 in spec2.locations:
      # unreachable rule
      for channel in inputs+outputs:
	t1_2 = pynta.Transition(source=loc_dict[(l1,l2)],target = universal)
	if channel in inputs:
	  t1_2.controllable = False
	  t1_2.synchronisation = (channel,"?")
	else: 
	  t1_2.controllable = True
	  t1_2.synchronisation = (channel,"!")
	if l1.invariant:
	  t1_2.guard = str(~l1.invariant_dbm)
	quotient.transitions.append(t1_2)
      # unsafe rule
      t1_2 = pynta.Transition(source=loc_dict[(l1,l2)],target = inconsistent, controllable=False) 	# inew transitions from th theory
      guard_dbm = ~l1.invariant_dbm & l2.invariant_dbm
      if not guard_dbm.isEmpty():
	t1_2.guard = str(guard_dbm)
      quotient.transitions.append(t1_2)
  
  # all rule
  for t1 in spec1.transitions:
    for t2 in spec2.transitions:
      if t1.synchronisation[0] == t2.synchronisation[0]:
	target = loc_dict[(t1.target,t2.target)]
	t1_2 = pynta.Transition(source=loc_dict[(t1.source,t2.source)], target = target)
	if t1.synchronisation[0] in inputs:
	  t1_2.controllable = False
	  t1_2.synchronisation = (t1.synchronisation[0],"?")
	else: 
	  t1_2.controllable = True
	  t1_2.synchronisation = (t1.synchronisation[0],"!")
	if t1.guard:
	  if t2.guard:
	    t1_2.guard = t1.guard + " and " + t2.guard
	  else:
	    t1_2.guard = t1.guard
	else:
	  t1_2.guard = t2.guard
	if t1.update:
	  if t2.update:
	    t1_2.update = t1.update + " , " + t2.update
	  else:
	    t1_2.update = t1.update
	else:
	  t1_2.update = t2.update
	quotient.transitions.append(t1_2)

  #unsafe
  for t2 in spec2.transitions:
    if t2.controllable:
      for l1 in spec1.locations:
	union_guards = spec1.context.getTautologyFederation()
	for t1 in l1.forward:
	  if t1.synchronisation[0] == t2.synchronisation[0]:
	    union_guards |= t1.guard_dbm
	t1_2 = pynta.Transition(source=loc_dict[(l1,t2.source)],target = inconsistent, controllable = False, synchronisation = (channel,"?"))
	union_guards = ~union_guards
	if not union_guards.isEmpty():
	  t1_2.guard = str(union_guards)
	quotient.transitions.append(t1_2)    

  #dividend
  for t1 in spec1.transitions:
    if t1.synchronisation[0] not in output2+inputs2:
      for l2 in spec2.locations:
	t1_2 = pynta.Transition(source = loc_dict[(t1.source,l2)],target = loc_dict[(t1.target,l2)], guard=t1.guard, update=t1.update)
	if t1.synchronisation[0] in inputs:
	  t1_2.controllable = False
	  t1_2.synchronisation = (t1.synchronisation[0],"?")
	else: 
	  t1_2.controllable = True
	  t1_2.synchronisation = (t1.synchronisation[0],"!")
	quotient.transitions.append(t1_2) 
  
  quotient.parse_template()
  makeInputEnabled(quotient,outputs,universal)
  quotient.update_forward_backward()
  return quotient
     
     

def invert(spec,outputs):
  """Invert a specification such that:
   - input transitions become output
   - output transitions become input
   - invariants are removed and added to the guards of incident edges
   - if the invariant is violated a transition is taken to an error location Unsafe
   - the model is made input enabled: any input not accepted previously is accepted by a transition to  the location UNSAFE
  """
  invert_spec = copy.deepcopy(spec)
  invert_spec.name += "Invert"
  for l in invert_spec.locations:
    l.template = invert_spec
  for t in invert_spec.transitions:
    t.template = invert_spec  
  unsafe_loc = pynta.Location(invert_spec,name="Unsafe")
  invert_spec.locations.append(unsafe_loc)
  
  for t in invert_spec.transitions:
    if not t.controllable: # input
      t.synchronisation = (t.synchronisation[0],'!')
      t.controllable = True
    else: # output
      t.synchronisation = (t.synchronisation[0],'?')
      t.controllable = False
    
    inv_str = str(t.source.invariant_dbm)
    if inv_str != "true":
      if str(t.guard):
	t.guard += " && " + inv_str[1:-1]
      else:
	t.guard = inv_str[1:-1]
    
  for l in invert_spec.locations:
    if l != unsafe_loc:
      inv_str = str(l.invariant_dbm)
      if inv_str != "true":
	invl = inv_str[1:-1].split("&")
	for invi in invl:
	  invert_spec.transitions.append(pynta.Transition(source=l, \
		  target = unsafe_loc, \
		  controllable = False, \
		  guard = contraryConstraint(invi), \
		  synchronisation = ("","?")))
	l.invariant = l.invariant_var
  invert_spec.parse_template()
  makeInputEnabled(invert_spec,outputs,unsafe_loc)
  invert_spec.update_forward_backward()  
  return invert_spec
  

def makeInputEnabled(spec,inputs,error_loc):
    """ Make a specification input enabled by linking all non accepted inputs to the error location"""
    #WARNING may not work with variables
    for l in spec.locations:
      for channel in inputs:
	conj_guards = []
	skip = False
	for t in spec.transitions:
	  if t.source == l and t.synchronisation[0] == channel:
	    guard_str = str(t.guard_dbm)
	    if guard_str == "true":
	      skip = True
	      break
	    else:
	      contrary_guard = []
	      guardl = guard_str[1:-1].split("&")
	      for guardi in guardl:
		contrary_guard.append(contraryConstraint(guardi))
	      conj_guards.append(contrary_guard)
	if not skip:
	  disj_guards= [[]]
	  for disj in conj_guards:
	    new_disj_guards = []
	    for conj in disj_guards:
	      for i in range(len(disj)-1):
		new_conj = copy.copy(conj)
		new_conj.append(disj[i])
		new_disj_guards.append(new_conj)
	      conj.append(disj[len(disj)-1])
	    disj_guards += new_disj_guards
	  for conj in disj_guards:
	    spec.transitions.append(pynta.Transition(source=l, \
		target = error_loc, \
		controllable = False, \
		guard = ' && '.join(conj), \
		synchronisation = (channel,"?")))
		
    
def removeUnreachableLocations(template):
    done = False
    while(not done):
      done = True
      locations_list = copy.copy(template.locations)
      for l in locations_list:
	isUnreachable = True
	for t in template.transitions:
	  if t.target == l and t.source != l:
	    isUnreachable = False
	    break
	if isUnreachable:
	  transitions_list = copy.copy(template.transitions)
	  template.locations.remove(l)
	  for t in transitions_list:
	    if t.target == l or t.source== l:
	      template.transitions.remove(t)
	  done = False
    
    
    
    
    