#ifndef PPL_TA_INT_H
#define PPL_TA_INT_H

#include <iostream>
#include <string>
#include <vector>

#include <ppl.hh>

using namespace std;
using namespace Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators;


/*! \brief An implementation of parametric federations for TA based on polyhedra using the PPL library.
 *  We assume the parameters are in the first variables of the domain. All others variables are clocks.
 */	
class ParamFederation : public Pointset_Powerset<NNC_Polyhedron>
{
  public:
    /*! \brief Empty constructor for an EMPTY domain. */	
    ParamFederation(int num_dimensions=0, Degenerate_Element kind=UNIVERSE) : Pointset_Powerset<NNC_Polyhedron>(num_dimensions,kind) { }
    
    /*! \brief Constructor with an initial polyhedron. */	
    ParamFederation(const NNC_Polyhedron& p) : Pointset_Powerset<NNC_Polyhedron>(p) { };
    /*! \brief Constructor from a linear expression and a coefficient.
     *  Type is :
     *  - 0 for EQUALITY (e == n)
     *  - 1 for NONSTRICT_INEQUALITY (e >= n)
     *  - 2 for STRICT_INEQUALITY (e > n)
     *  
     */	
    ParamFederation(int num_dimensions, const Linear_Expression& e, const int n, const int type); 
    /*! \brief Constructor from two linear expressions.
     *  Type is :
     *  - 0 for EQUALITY (e1 == e2)
     *  - 1 for NONSTRICT_INEQUALITY (e1 >= e2)
     *  - 2 for STRICT_INEQUALITY (e1 > e2)
     *  
     */	
    ParamFederation(int num_dimensions, const Linear_Expression& e1, const Linear_Expression& e2, const int type);
    /*! \brief Copy constructor. */	
    ParamFederation(const ParamFederation& fed) : Pointset_Powerset<NNC_Polyhedron>(fed) { }

    /*! \brief Number of dimensions. */	
    int size() const {
      return this->space_dimension();
    }
    
    //////////// Add new disjunctions:
    
    /*! \brief Add a constraint in disjunction. */	
    void add_disjunction(const Constraint& c);

    /*! \brief Add a formula in disjunction. 
      *  Add each polyhedron in fed in disjunction.
      */	
    void add_disjunction(const ParamFederation& fed);
    
    //////////// Add in conjunction:

    /*! \brief Add a constraint in conjunction. */	
    void add_conjunction(const Constraint& c);

    /*! \brief Add a formula in conjunction. */
    void add_conjunction(const ParamFederation& fed);
    
    ///////////// Negations
    /*! \brief Compute the negation of this formula. */
    ParamFederation negation() const;
    
    /*! \brief Compute the negation of a single constraint. */	
    static Constraint negation(const Constraint& c);

   
    /*! \brief Assign to this the difference between this and the federation passed in parameter.
      *  P\Q = P and not Q
      */
    void difference(const ParamFederation& fed);
    
    
    ////////////// Timed Automata specific 
    /*! \brief Assign to this the future of this.
     * The parameters in the first num_param variables are constant in this transformation.
     */
    void future(int num_param);
    
    /*! \brief Assign to this the past of this. 
     * The parameters in the first num_param variables are constant in this transformation.
     */
    void past(int num_param);
    
    /*! \brief Free a variable to all positive values. */
    void free(int var);
    
    /*! \brief Reset to zero the variable given in parameter. */
    void reset(int var);    

    /*! \brief Compute and assign to this the post states of this domain through an edge with guard and reset cloks. */
    void post(const ParamFederation& guard, const std::vector<int>& reset);

    /*! \brief Compute and assign to this the pre states of this domain through an edge with guard and reset cloks. */
    void pred(const ParamFederation& guard, const std::vector<int>& reset);
    
    /*! \brief Compute the time predecessors of good that avoid bad. 
     *  Good and bad are federations. Compute a union of intersection of predt applied between two polyhedra in the federation.
     */
    static ParamFederation predt(const ParamFederation& good,const ParamFederation& bad,int num_param);
    
    ////////////// Parameter specific 
        
    /*! \brief Projection on the set of parameters. */
    void param_projection(int num_param);
    
    /*! \brief Return the infinimum value of the parameter over the domain.
     *  - id_param is the variable index of the parameter 
     *  - min_n is the result numerator
     *  - min_d is the result denumerator
     *  - bool is true if the result is also a minimum value
     *  - found is true if a infinimum was found
     */
    void minimize_param(int id_param, bool* found, int* min_n, int* min_d, bool* minimum) const;
    
    ////////////// Printing
    /*! \brief Print the domain intro a string. */
    string toStr() const;
   
    
  private:
    /*! \brief Compute the time predecessors of good that avoid bad.
     *  Good and bad are convex polyhedron.
     */
    static ParamFederation predt(const NNC_Polyhedron& good,const NNC_Polyhedron& bad,int num_param);
    
};

#endif

