"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""


import pynta
import pyparamnta
import pyrobustcegar
import pytiga
import pyreachability
import pytadactl
import model_transformations
import misc_functions

from fractions import Fraction

import logging
import copy

## configuration of the logging system
logging.basicConfig()
logger = logging.getLogger('ECDAR')

logger.setLevel(logging.INFO)
pynta.logger.setLevel(logging.INFO)
pyparamnta.logger.setLevel(logging.INFO)
pyrobustcegar.logger.setLevel(logging.INFO)
pytiga.logger.setLevel(logging.INFO)
pyreachability.logger.setLevel(logging.INFO)
pytadactl.logger.setLevel(logging.INFO)
model_transformations.logger.setLevel(logging.INFO)
misc_functions.logger.setLevel(logging.INFO)


def loadModel(filename):
  """ Loads a XML file with a network of timed automata into a workspace"""
  try:
    nta = pynta.NTA.from_xml(filename)
  except IOError as (errno, strerror):
    logger.error("I/O error({0}): {1}".format(errno, strerror))
  else:
    return Workspace(nta,filename)


class Workspace:
  """ Container for a network of timed automata"""
  def __init__(self,nta,filename):
    self.nta=nta
    self.filename = filename
    
  def addTemplate(self,template_model):
    """template_model is either a game or a specification"""
    if isinstance(template_model,Specification):
      self.nta.templates.append(template_model.template)
    elif isinstance(template_model,pynta.Template):
      self.nta.templates.append(template_model)
    elif isinstance(template_model,pytiga.TimedGame):
      self.nta.templates.append(template_model.goal_template)
    elif isinstance(template_model,pytiga.SafetyTimedGame):
      self.nta.templates.append(template_model.goal_template)
    
  def save(self,filename=""):
    if not filename:
      filename = self.filename
    try:
      f = open(filename, 'w')
      f.write(self.nta.to_xml())
      f.close()
    except IOError as (errno, strerror):
      logger.error("I/O error({0}): {1}".format(errno, strerror))

  def getSpecification(self,name):
    ts = [t for t in self.nta.templates if t.name == name] 
    if len(ts) != 1:
      logger.error("No or more than one template with name \""+name+"\"")
      return None
    else:
      return Specification(ts[0])
  
  def getTimedGame(self,template,objective):
    if isinstance(template,str):
      ts = [t for t in self.nta.templates if t.name == template] 
      if len(ts) != 1:
	logger.error("No or more than one template with name \""+name+"\"")
	return None
      else:
	template = copy.deepcopy(ts[0])
    return pytiga.TimedGame(template,objective)
    
    
   

class Specification:
  def __init__(self,template):
    self.template = template
    self.inputs=[]
    self.outputs=[]
    self.template.parse_template()
    self.universal = self.template.get_location("Universal")
    self.inconsistent = self.template.get_location("Inconsistent")
    for transition in self.template.transitions:
      if transition.synchronisation:
	channel_name = transition.synchronisation[0]
	sync_type = transition.synchronisation[1]
	if sync_type is '!':
	  if channel_name not in self.outputs:
	    self.outputs.append(channel_name)
	elif sync_type is '?':
	  if channel_name not in self.inputs:
	    self.inputs.append(channel_name)
	else:
	  logger.warning("Unknown synchronisation type: " + sync_type)
      else:
	logger.warning("Transition with no synchronisation.")
  
  def isSpecification(self):
    """Check if the template syntax correspond to a specification:
    - no features
    - all transition are synchronized
    - disjoint input and output
    - reset values are zero (always the case in the current version)
    - deterministic
    - input enabled
    - all the clocks are local (always the case in the current version)
    """
    if self.template.get_nb_features() > 0:
      logger.error("the model has features.")
      return False
    for transition in self.template.transitions:
      if not transition.synchronisation:
	logger.error("transition \"" + str(transition) + "\" has no synchronisation.")
	return False
      if transition.synchronisation[0] in self.inputs and transition.synchronisation[0] in self.outputs:
	logger.error("inputs and outputs channel are not disjoint.")
	return False
    if not self.isDeterministic():
      logger.error("the model is not deterministic.")
      return False
    if not self.isInputEnabled():
      logger.error("the model is not input-enabled.")
      return False
    return True
      
  def isDeterministic(self):
    for state in self.template.get_list_discrete_states():
      for index_t1 in range(0,len(state.location.forward)):
	t1 = state.location.forward[index_t1]
	if t1.has_next(state):
	  for index_t2 in range(index_t1+1,len(state.location.forward)):
	    t2 = state.location.forward[index_t2]
	    if t1.synchronisation ==  t2.synchronisation and t2.has_next(state):
	      dbm = t1.guard_dbm & t2.guard_dbm & state.location.invariant_dbm
	      if not dbm.isEmpty():
		logger.error("Discrete state " + state.to_str() + " is not deterministic with respect to action \"" + t1.synchronisation[0] + "\"")
		return False
    return True
    
  def isInputEnabled(self):
    for state in self.template.get_list_discrete_states():
      for i in self.inputs:
	dbm = self.template.context.getEmptyFederation()
	for t in state.location.forward:
	  if t.synchronisation[0] == i and t.has_next(state):
	    dbm |= t.guard_dbm
	dbm &= state.location.invariant_dbm
	if dbm != state.location.invariant_dbm:
	  logger.error("Discrete state " + state.to_str() + " is not input enabled with respect to action \"" + i + "\"")
	  return False
    return True

  #def isImplementation(self):
    #"""Check if the template syntax correspond to an implementation automata:
    #- output urgency (except for the universal location)
    #- independent progress
    #"""
    #if not self.isSpecification():
      #return False
  
  def scale(self,factor):
    return Specification(model_transformations.scaledTemplateConstants(self.template,factor))
  
  
  ### Consistency checking  ----------------------------------------------------------------------------------------
  
  def getConsistencyGame(self):
    cons_game = model_transformations.consistencyGame(self.template)
    cons_game.parse_template()
    return pytiga.SafetyTimedGame(cons_game,["Inconsistent"])
    
  def getRobustConsistencyGame(self,delta="delta",param=0):
    """Construct the robust consistency game. 
      - delta is either the value of delta, or the name of the parameter (by default: delta)
      - param is the value of delta as a parameter (by default 0)
      Note that if delta is a value, param is useless.
      Return the template corresponding to the game and if delta is a parameter, add a const declaration in the nta.
      """
    if isinstance(delta,str):
      cons_game = model_transformations.consistencyGame(self.template,delta)
      return pytiga.SafetyTimedGame(model_transformations.robustTGA(cons_game,delta,"Inconsistent",param=0),["Inconsistent"])
    else:
      try:
	fdelta = Fraction(str(delta))
      except:
	logger.error("invalid input parameter: "+ str(delta))
	raise
      scaled_game = model_transformations.scaledTemplateConstants(self.template,fdelta.denominator)
      scaled_game.parse_template()
      cons_game = model_transformations.consistencyGame(scaled_game,fdelta.numerator)
      cons_game.parse_template()
      robust_cons_game = model_transformations.robustTGA(cons_game,fdelta.numerator,"Inconsistent")
      robust_cons_game.parse_template()
      return pytiga.SafetyTimedGame(robust_cons_game,["Inconsistent"])  
    
  def isConsistent(self):
    cons_game = model_transformations.consistencyGame(self.template)
    cons_game.parse_template()
    timedGame = pytiga.SafetyTimedGame(cons_game,["Inconsistent"])
    return timedGame.solve()
  
  def isRobustConsistent(self,delta):
    """Delta is a decimal value"""
    try:
      fdelta = Fraction(str(delta))
    except:
      logger.error("invalid input parameter: "+ str(delta))
      raise
    scaled_game = model_transformations.scaledTemplateConstants(self.template,fdelta.denominator)
    scaled_game.parse_template()
    cons_game = model_transformations.consistencyGame(scaled_game,fdelta.numerator)
    cons_game.parse_template()
    robust_cons_game = model_transformations.robustTGA(cons_game,fdelta.numerator,"Inconsistent")
    robust_cons_game.parse_template()
    timedGame = pytiga.SafetyTimedGame(robust_cons_game,["Inconsistent"])
    return timedGame.solve()
  
    
  def maxRobustConsistent(self,delta_max,confidence):
    """ delta_max is an integer
        confidence is a decimal"""
    cons_game = model_transformations.consistencyGame(self.template,"delta")
    cons_game.parse_template()
    robust_cons_game = model_transformations.robustTGA(cons_game,"delta","Inconsistent",param=delta_max)
    robust_cons_game.parse_template()
    cegarAlgo = pyrobustcegar.CegarForRobustness(robust_cons_game,["Inconsistent"],2,str(confidence),delta_max)
    return cegarAlgo.cegar()
    
  def maxRobustConsistentBS(self,delta_max,confidence):
    """ delta_max is an integer
        confidence is a decimal"""
    cons_game = model_transformations.consistencyGame(self.template,"delta")
    cons_game.parse_template()
    robust_cons_game = model_transformations.robustTGA(cons_game,"delta","Inconsistent",param=delta_max)
    robust_cons_game.parse_template()
    cegarAlgo = pyrobustcegar.CegarForRobustness(robust_cons_game,["Inconsistent"],2,str(confidence),delta_max)
    return cegarAlgo.binary_refine()
    
    
  ### Usefulness checking --------------------------------------------------------------------------------------------------
  
  def getUsefulnessGame(self,bad_loc_list=[]):
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    useful_game.parse_template()
    return pytiga.SafetyTimedGame(useful_game,"Universal")
    
  def getRobustUsefulnessGame(self,delta="delta",param=0,bad_loc_list=[]):
    if isinstance(delta,str):
      useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list=[])
      return pytiga.SafetyTimedGame(model_transformations.robustTGA(useful_game,delta,"Universal",param),"Universal")
    else:
      try:
	fdelta = Fraction(str(delta))
      except:
	logger.error("invalid input parameter: "+ str(delta))
	raise
      useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
      useful_game.parse_template()
      scaled_game = model_transformations.scaledTemplateConstants(useful_game,fdelta.denominator)
      scaled_game.parse_template()
      robust_useful_game = model_transformations.robustTGA(scaled_game,fdelta.numerator,"Universal")
      robust_useful_game.parse_template()
      return pytiga.SafetyTimedGame(robust_useful_game,"Universal")
    
  def isUseful(self,bad_loc_list=[]):
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    useful_game.parse_template()
    timeGame = pytiga.SafetyTimedGame(useful_game,"Universal")
    return timeGame.solve()
    
  def isRobustUseful(self,delta,bad_loc_list=[]):
    """Delta is a decimal value"""
    try:
      fdelta = Fraction(str(delta))
    except:
      logger.error("invalid input parameter: "+ str(delta))
      raise    
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    useful_game.parse_template()
    scaled_game = model_transformations.scaledTemplateConstants(useful_game,fdelta.denominator)
    scaled_game.parse_template()
    robust_useful_game = model_transformations.robustTGA(scaled_game,fdelta.numerator,"Universal")
    robust_useful_game.parse_template()
    timedGame = pytiga.SafetyTimedGame(robust_useful_game,"Universal")
    return timedGame.solve()
    
  def maxRobustUseful(self,delta_max,confidence,bad_loc_list=[]):
    """ delta_max is an integer
        confidence is a decimal"""
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    useful_game.parse_template()
    robust_useful_game = model_transformations.robustTGA(useful_game,"delta","Universal",param=delta_max)
    robust_useful_game.parse_template()
    cegarAlgo = pyrobustcegar.CegarForRobustness(robust_useful_game,"Universal",2,str(confidence),delta_max)
    return cegarAlgo.cegar()
    
  def maxRobustUsefulBS(self,delta_max,confidence,bad_loc_list=[]):
    """ delta_max is an integer
        confidence is a decimal"""
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    useful_game.parse_template()
    robust_useful_game = model_transformations.robustTGA(useful_game,"delta","Universal",param=delta_max)
    robust_useful_game.parse_template()
    cegarAlgo = pyrobustcegar.CegarForRobustness(robust_useful_game,"Universal",2,str(confidence),delta_max)
    return cegarAlgo.binary_refine()
    
  ### Refinement -------------------------------------------------------------------------------------------------------------------------
  
  def refine(self,big_spec):
    invert_big_spec_template =  model_transformations.invert(big_spec.template,big_spec.outputs)
    invert_big_spec_template.parse_template()
    check_template = model_transformations.composition(self.template,invert_big_spec_template, \
			    self.outputs, big_spec.inputs, self.inputs, big_spec.outputs, None, None)
    check_template.parse_template()
    bad_locs = []
    for l in check_template.locations:
	l_split = l.name.split("PAR")
	if ("Unsafe" in l_split):
	    bad_locs.append(l.name)
    refCheck = pyreachability.ReachabilityAnalysis(check_template,bad_locs,True)
    return refCheck.solve()
   
  def robSat(self,big_spec,delta):
    try:
      fdelta = Fraction(str(delta))
    except:
      logger.error("invalid input parameter: "+ str(delta))
      raise    
    scaled_spec = big_spec.scale(fdelta.denominator)
    perturbed_impl = self.implemPerturbation(delta)
    return perturbed_impl <= scaled_spec
  
  def maxRobSat(self,big_spec,delta_max,confidence):
    """ delta_max is an integer
        confidence is a decimal"""
    perturbed = self.implemPerturbation("delta")  #("1*delta")
    check_template = model_transformations.composition(perturbed.template,model_transformations.invert(big_spec.template,big_spec.outputs), \
			    self.outputs, big_spec.inputs, self.inputs, big_spec.outputs, None, None)
    check_template.parse_template()
    bad_locs = []
    for l in check_template.locations:
	l_split = l.name.split("PAR")
	if ("Unsafe" in l_split):
	    bad_locs.append(l.name)
    cegarAlgo = pyrobustcegar.CegarForRobustness(model_template,bad_locs,1,str(confidence),delta_max)
    return cegarAlgo.cegar()
    
  def maxRobSatBS(self,big_spec,delta_max,confidence):
    """ delta_max is an integer
        confidence is a decimal"""
    perturbed = self.implemPerturbation("delta")  #("1*delta")
    check_template = model_transformations.composition(perturbed.template,model_transformations.invert(big_spec.template,big_spec.outputs), \
			    self.outputs, big_spec.inputs, self.inputs, big_spec.outputs, None, None)
    check_template.parse_template()
    bad_locs = []
    for l in check_template.locations:
	l_split = l.name.split("PAR")
	if ("Unsafe" in l_split):
	    bad_locs.append(l.name)			    
    cegarAlgo = pyrobustcegar.CegarForRobustness(model_template,bad_locs,1,str(confidence),delta_max)
    return cegarAlgo.binary_refine()
    
  def __le__(self,big_spec):
    return self.refine(big_spec)
    
  def __ge__(self,small_spec):
    return small_spec.refine(self)
    
  def implemPerturbation(self,delta="delta",param=0):
    """Construct the delta-perturbation of an implementation.
      - delta is either the value of delta, or the name of the parameter (by default: delta)
      - param is the value of delta as a parameter (by default 0)
      Note that if delta is a value, param is useless.
      Return a new template corresponding, and if delta is a parameter, add a const declaration in the nta.
      """
    if isinstance(delta,str):
      return model_transformations.implemPerturbation(self.template,delta,self.outputs,self.inputs,param)
    else:
      try:
	fdelta = Fraction(str(delta))
      except:
	logger.error("invalid input parameter: "+ str(delta))
	raise
      scaled_impl = model_transformations.scaledTemplateConstants(self.template,fdelta.denominator)
      scaled_impl.parse_template()
      return Specification(model_transformations.implemPerturbation(scaled_impl,fdelta.numerator,self.outputs,self.inputs))
  
    
  ### Conjunction and composition -------------------------------------------------------------------------------------------------------------------------

  def sameAlphabet(self,spec2):
    """ Check if two specifications have the same action alphabet.
	This condition is required to compute the conjunction.  """
    for a in self.inputs:
      if a not in spec2.inputs:
	return False
    for a in self.outputs:
      if a not in spec2.outputs:
	return False
    return True
    
  def composable(self,spec2):
    """Check that the two specifications are composable: their output alphabets are disjoint"""
    for a in self.outputs:
      if a in spec2.outputs:
	return False
    return True  
  
  def quotientable(self,spec2):
    """	Check Checks if the specification can be quotiented by spec2
	This requires that the inputs of spec2 are included in the inputs of the specifications,
	and the outputs of spec2 are included in the outputs of the specifications.
    """
    for a in spec2.inputs:
      if a not in self.inputs:
	return False
    for a in spec2.outputs:
      if a not in self.outputs:
	return False
    return True
  
  def conjunction(self,spec2,merge_universal=True):
    if not self.sameAlphabet(spec2):
      logger.error("the specifications have two different alphabet.")
      return None
    return Specification(model_transformations.conjunction(self.template,spec2.template,self.outputs,self.inputs,self.universal,spec2.universal,merge_universal))
  
  def composition(self,spec2,merge_universal=True):
    if not self.composable(spec2):
      logger.error("the specifications are not composable.")
      return None
    return Specification(model_transformations.composition(self.template,spec2.template,self.outputs,spec2.outputs,self.inputs,spec2.inputs,self.universal,spec2.universal,merge_universal))

  def quotient(self,spec2,):
    if not self.quotientable(spec2):
      logger.error("the specifications cannot be quotiented.")
      return None
    return Specification(model_transformations.quotient(self.template,self.outputs,spec2.outputs,self.inputs,spec2.inputs))
  
  def __and__(self,spec2):
    return self.conjunction(spec2)
    
  def __or__(self,spec2):
    return self.composition(spec2)
  
  def __div__(self,spec2):
    return self.quotient(spec2)

