"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import pytiga
import pyreachability
import pyparamnta

import logging
import sys

from model_transformations import scaledTemplateConstants
from fractions import Fraction

logging.basicConfig()
logger = logging.getLogger('CEGAR')

class CegarForRobustness:
  def __init__(self, goal_template, goal_locations_name, problem, confidence, max_delta_int):
    assert(isinstance(goal_template,pyparamnta.ParamTemplate))
    self.goal_template = goal_template
    self.goal_locations_name = goal_locations_name
    self.problem = problem			# 1 for 1-player safety, 2 for 2-player safety 

    self.max_delta = Fraction(max_delta_int,1)  # max_delta is a infinimum of the incorrect values for the parameter delta in the robustness problem
    self.max_is_bad = False 			# true if max_delta is also a minimum (i.e. is an incorrect value)  
    self.min_delta = Fraction(0,1)		# min_delta is a maximum of the correct values
    self.confidence = Fraction(confidence)

  def choose_new_delta(self,last_delta):
    if last_delta >= self.max_delta:
      if not self.max_is_bad and last_delta-self.max_delta>=self.confidence:
	return self.max_delta
      else:
	ret = self.max_delta-self.confidence
	if ret > 0:
	  return ret
	else:
	  return 0 
    else:
      logger.error("Monotonicity hypothesis is broken: delta value is increasing.")
      return -1
    
  def cegar(self):
    current_delta = self.max_delta+self.confidence
    while self.max_delta-self.min_delta > self.confidence:
      current_delta = self.choose_new_delta(current_delta)
      if current_delta < 0:
	self.min_delta = -1
	break
      logger.info(" New game with value " + str(current_delta))
      if self.refine_game(current_delta) < 0:
	break
      logger.info(" ...refinement result: max=" + str(self.max_delta) + " min=" + str(self.min_delta) + " strict: " + str(self.max_is_bad))
    return float(self.min_delta)
    
  def refine_game(self,current_delta):
    scaled_template = scaledTemplateConstants(self.goal_template,current_delta.denominator)
    scaled_template.setParamValue(current_delta.numerator)
    scaled_template.parse_template()
    scaled_template.parse_param(0,current_delta.numerator)
    if self.problem == 1:
      game = pyreachability.ReachabilityAnalysis(scaled_template, self.goal_locations_name, True)
    elif self.problem == 2:
      game = pytiga.SafetyTimedGame(scaled_template, self.goal_locations_name)
    else:
      logger.error("Unknown value for problem.")
      return -1
    if game.solve():
      logger.info(" ...game is won;")
      self.min_delta = current_delta
    else:
      logger.info(" ...game is lost; refining...")
      #game.extractStrategy()
      result = game.refine_safety(current_delta.numerator)
      if result:
	max_n,max_d,is_max = result
      else:
	logger.error("Unable to refine the counter strategy.")
	sys.exit(1)
      self.max_delta = Fraction(max_n,max_d*current_delta.denominator)
      self.max_is_bad = is_max
    return self.min_delta
 
 
  def binary_refine(self):
    while self.max_delta-self.min_delta > self.confidence:
      current_delta = (self.max_delta-self.min_delta)/2 + self.min_delta
      logger.info(" New game with value " + str(current_delta))
      
      scaled_template = scaledTemplateConstants(self.goal_template,current_delta.denominator)
      scaled_template.setParamValue(current_delta.numerator)
      scaled_template.parse_template()
      if self.problem == 1:
	game = pyreachability.ReachabilityAnalysis(scaled_template, self.goal_locations_name, True)
      elif self.problem == 2:
	game = pytiga.SafetyTimedGame(scaled_template, self.goal_locations_name)
      else:
	logger.error("Unknown value for problem.")
	return -1
      if game.solve():
	logger.info(" ...game is won;")
	self.min_delta = current_delta
      else:
	logger.info(" ...game is lost;")
	self.max_delta = current_delta
	self.max_is_bad = True
      logger.info(" ...refinement result: max=" + str(self.max_delta) + " min=" + str(self.min_delta) + " strict: " + str(self.max_is_bad))
    return float(self.min_delta)
    

      
  
  
    
    
  
