"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import logging
import copy

import pycudd

from misc_functions import evaluate_expression

logging.basicConfig()
logger = logging.getLogger('PYSYMSTATE')
  
class SymbolicState:
  """A symbolic state composed with:
   - a zone
   - a bdd (optionally: it may be None)
  """
  
  def __init__(self,bdd,zone):
    self.bdd = bdd
    self.zone = zone
  
  def copy(self):
    ret = copy.copy(self)
    ret.bdd = copy.copy(self.bdd)
    ret.zone = self.zone.copy()         
    return ret
  
  def to_str(self,template):
    if self.bdd:
      return "("+  template.bdd_to_str(self.bdd) + ", " + str(self.zone) + ")"
    else:
      return "("+  str(self.zone) + ")"
  
class SetOfSymbolicStates(list):
  """A union of symbolic states with a list of values of variables
  """
  
  def __init__(self,template):
    self.template = template
  
  def copy(self):
    ret = self.__class__(self.template)
    for state in self:
      ret += [state.copy()]
    return ret
  
  def __str__(self):
    ret = ""
    for state in self:
      if ret:
	ret += ", "
      ret += state.to_str(self.template)
    return "[" + ret + "]"
  
  def is_included_in(self,other):
    #DBM inclusion only
    self_zone = self.template.context.getEmptyFederation()
    for state in self:
      self_zone |= state.zone    
    other_zone = self.template.context.getEmptyFederation()
    for state in other:
      other_zone |= state.zone
    return self_zone <= other_zone
  
  def __lt__(self,other):
    return self.is_included_in(other)
  
  def __gt__(self,other):
    return other.is_included_in(self)
  
  def append(self,new_state):
    """ Add a new symbolic state.
	Apply reduction rules to merge with the states already in.
	Modify new_state and return the part of the symbolic state that is new in the list.
    """   
    if not self.template.mgr:
      if self:
	assert len(self)==1
	self[0].zone |= new_state.zone
      else:
	self += [new_state]
      return new_state
    if (new_state.bdd.Empty() | new_state.zone.isEmpty()):
      return None
    index = 0
    equal_state = None
    for state in self:
      if new_state.zone <= state.zone:
	new_state.bdd = new_state.bdd & ~state.bdd
	if new_state.bdd.Empty():
	  return None
	if new_state.zone == state.zone:
	  #assert(not equal_state)  #only one zone may be equal because otherwise the two already present should have been merged
	  equal_state = state
      elif state.zone < new_state.zone:
	state.bdd = state.bdd & ~new_state.bdd
	if state.bdd.Empty():
	  self.pop(index)
      index += 1
    if equal_state:
      #new_state.bdd = new_state.bdd & ~equal_state.bdd # already done 14 lines above
      equal_state.bdd = equal_state.bdd | new_state.bdd
      if new_state.bdd.Empty():
	return None
    else:
      self += [new_state]
    return new_state
  
  def __invert__(self):
    result = SetOfSymbolicStates(self.template)
    if not self.template.mgr:
      if self:
	result += [SymbolicState(None,~self[0].zone)]
      else:
	result += [SymbolicState(None,self.template.context.getTautologyFederation())]
      return result
    if not self:
      result += [SymbolicState(self.template.mgr.One(),self.template.context.getTautologyFederation())]
      return result
    current_state = self[0]
    result.append(SymbolicState(current_state.bdd,~current_state.zone))
    result.append(SymbolicState(~current_state.bdd,self.template.context.getTautologyFederation()))
    for index in range(1,len(self)):
      current_state = self[index]
      old_results = result
      result = SetOfSymbolicStates(self.template)
      for state in old_results:
	result.append(SymbolicState(state.bdd & current_state.bdd,state.zone & ~current_state.zone))
	result.append(SymbolicState(state.bdd & ~current_state.bdd,state.zone))
    return result
  
  def __or__(self,arg):
    result = self.copy()
    for state in arg:
      result.append(state.copy())
    return result
  
  def __ior__(self,arg):
    for state in arg:
      self.append(state.copy())
    return self
    
  def __and__(self,arg):
    result = SetOfSymbolicStates(self.template)
    if not self.template.mgr:
      if self and arg:
	result += [SymbolicState(None,self[0].zone & arg[0].zone)]
      return result
    
    for state_self in self:
      for state_arg in arg:
	result.append(SymbolicState(state_self.bdd & state_arg.bdd, state_self.zone & state_arg.zone))
    return result
  
  def __iand__(self,arg):
    if not self.template.mgr:
      if self and arg:
	self[0].zone &= arg[0].zone
      else:
	self = SetOfSymbolicStates(self.template)
      return self
      
    result = SetOfSymbolicStates(self.template)
    for state_self in self:
      for state_arg in arg:
	result.append(SymbolicState(state_self.bdd & state_arg.bdd, state_self.zone & state_arg.zone))
    self = result
    return self
  
  def __sub__(self, arg):
    result = self.copy()
    result &= ~arg
    return result
  
  def __isub__(self, arg):
    self &= ~arg
    return self

  def conjunction(self,state):
    result = SetOfSymbolicStates(self.template)
    if not self.template.mgr:
      if self:
	result = [SymbolicState(None,self[0].zone & state.zone)]
      return result
      
    for current_state in self:
      result.append(SymbolicState(current_state.bdd & state.bdd, current_state.zone & state.zone))
    return result
  
  def conjunction_in_place(self,state):
    if not self.template.mgr:
      if self:
	self[0].zone &= state.zone
      else:
	self = SetOfSymbolicStates(self.template)
      return self
      
    for index in xrange(len(self) - 1, -1, -1):
      current_state = self[index]
      current_state.bdd &= state.bdd
      current_state.zone &= state.zone
      if (current_state.bdd.Empty() | current_state.zone.isEmpty()):
	del self[index]
    return self
    
  def difference(self,big_state):
    neg = ~self
    neg.conjunction_in_place(big_state)
    return neg
 
  def freeClockInPlace(self,clock):
    for state in self:    
      state.zone.freeClockInPlace(clock)
    return self
    
  def up(self):
    result = self.copy()
    for state in result:
      state.zone.upInPlace()
    return result
  
  def upInPlace(self):
    for state in self:
      state.zone.upInPlace()
    return self
    
  def hasZero(self):
    for state in self:
      if state.zone.hasZero():
	return True
    return False
    
    
  def post(self, transition):
    ret_states = SetOfSymbolicStates(self.template)
    for state in self:
      ret_state = state.copy()
      if self.template.mgr:
	ret_state.bdd &= transition.guard_bdd
	if ret_state.bdd.Empty():
	  continue
      ret_state.zone &= transition.guard_dbm
      if ret_state.zone.isEmpty():
	continue
      
      # compute the next zone
      for clock in transition.update_clock:
	ret_state.zone.resetValueInPlace(clock)
      ret_state.zone &= transition.target.invariant_dbm
      if ret_state.zone.isEmpty():
        continue
      if not transition.target.urgent:
	ret_state.zone.upInPlace()
      ret_state.zone &= transition.target.invariant_dbm
      ret_state.zone.extrapolateMaxBoundsInPlace(self.template.clock_bounds)

      #compute the next bdd
      if self.template.mgr:
	adaptive_features = self.template.fa if transition.controllable else self.template.fe
	new_adaptive_features = self.template.fb if transition.controllable else self.template.ff
	swap_old = pycudd.DdArray(len(adaptive_features))
	swap_new = pycudd.DdArray(len(adaptive_features))
	for f in adaptive_features:  
	  swap_old.Push(self.template.features[f])
	for f in new_adaptive_features:
	  swap_new.Push(self.template.features[f])
	ret_state.bdd = ret_state.bdd.SwapVariables(swap_old,swap_new,len(adaptive_features))
	#print "swap variables : " + self.template.bdd_to_str(ret_state.bdd)
	for f in new_adaptive_features:
	  ret_state.bdd = ret_state.bdd.ExistAbstract(self.template.features[f])
	  #print "abstract adaptive feature " + str(f+self.template.get_nb_adaptive_features()) + " : " + self.template.bdd_to_str(ret_state.bdd)
	ret_state.bdd &= transition.source.invariant_bdd
  
      ret_states.append(ret_state)
    return ret_states
  
  def pred(self, transition):
    #print "computing predecessors of " + str(self)

    # computes the predecessors of this through transition
    ret_states = SetOfSymbolicStates(self.template)
    for state in self:    
      #print "computing predecessors of state " + state.to_str(self.template)

      ret_state = state.copy()
      if self.template.mgr:
	ret_state.bdd &= transition.target.invariant_bdd
      ret_state.zone &= transition.target.invariant_dbm
      
      # compute the preimage of the zone
      for clock in transition.update_clock:
	ret_state.zone &= (clock == 0)
	ret_state.zone.freeClockInPlace(clock)
      ret_state.zone &= transition.guard_dbm
      ret_state.zone &= transition.source.invariant_dbm

      #print "zone predecessors : " + str(ret_state.zone)

      # compute the preimage of the bdd
      if self.template.mgr:
	adaptive_features = self.template.fa if transition.controllable else self.template.fe
	new_adaptive_features = self.template.fb if transition.controllable else self.template.ff
	swap_old = pycudd.DdArray(len(adaptive_features))
	swap_new = pycudd.DdArray(len(new_adaptive_features))
	for f in adaptive_features:  
	  swap_old.Push(self.template.features[f])
	for f in new_adaptive_features:
	  swap_new.Push(self.template.features[f])
	ret_state.bdd = ret_state.bdd.SwapVariables(swap_old,swap_new,len(adaptive_features))
	#print "swap variables : " + self.template.bdd_to_str(ret_state.bdd)
	ret_state.bdd &= transition.guard_bdd
	#print "add guard : " + self.template.bdd_to_str(ret_state.bdd)
	for f in new_adaptive_features:  
	  ret_state.bdd = ret_state.bdd.ExistAbstract(self.template.features[f])
	  #print "abstract adaptive feature : " + self.template.bdd_to_str(ret_state.bdd)
	ret_state.bdd &= transition.source.invariant_bdd
	#print "bdd predecessors : " + self.template.bdd_to_str(ret_state.bdd)
      ret_states.append(ret_state)
    return ret_states
    
  def predt(self, location, bad_states):
    # computes the time predecessors of this that avoid bad_states
    if location.urgent:
      return self - bad_states
    ret_states = SetOfSymbolicStates(self.template)
    for good_state in self:
      if not self.template.mgr:
	if bad_states:
	  ret_states += [SymbolicState(None, good_state.zone.predt(bad_states[0].zone))]
	else:
	  ret_states += [SymbolicState(None, good_state.zone.predt(self.template.context.getEmptyFederation()))]
      else:
	pred_good_zone = good_state.zone.predt(self.template.context.getEmptyFederation())
	current_preds = SetOfSymbolicStates(self.template)
	current_preds.append(SymbolicState(good_state.bdd, good_state.zone.predt(self.template.context.getEmptyFederation())))
	
	for bad_state in bad_states:
	  bad_state_preds = SetOfSymbolicStates(self.template)
	  bad_state_pred1 = SymbolicState(~bad_state.bdd,self.template.context.getTautologyFederation())
	  if not bad_state_pred1.bdd.Empty():
	    bad_state_preds += [bad_state_pred1]
	  bad_state_pred2 = SymbolicState(bad_state.bdd, good_state.zone.predt(bad_state.zone))
	  if not bad_state_pred2.bdd.Empty() and not bad_state_pred2.zone.isEmpty():
	    bad_state_preds += [bad_state_pred2]
	  current_preds &= bad_state_preds
	ret_states |= current_preds
    return ret_states
    
    
class DiscreteState:
  """ The combination of a location and a set of values for the variables of the system (NTA variables + template variables)."""
  
  def __init__(self,location,variables):
    self.location = location
    self.variables = tuple(variables)
    
  def __hash__(self):
    if self.variables:
      return hash(self.location) + hash(self.variables)
    else:
      return hash(self.location)
  def __eq__(self, arg2):
    if self.variables:
      return self.location == arg2.location and self.variables == arg2.variables
    else:
      return self.location == arg2.location
  
  def __str__(self):
    ret = "Location \'" + self.location.name + "\'"
    if self.variables:
      ret += " with "
      for index in range(len(self.variables)):
	if index > 0:
	  ret += ","
	ret += "\'" + self.location.template.global_variables_names[index] + "\'=" + str(self.variables[index])
    return ret
  
  def evaluate_expression(self,constraints):
    variables = dict(zip(self.location.template.variables.items(), self.variables) +  self.location.template.nta.constants().items() +  self.template.nta.constants().items())
    val = evaluate_expression(constraints,variables)
    return val